// -------------------------------------------------------------------------
// $Id$
//  spiNNaker link driver module
//
// -------------------------------------------------------------------------
// AUTHOR
//  lap - luis.plana@manchester.ac.uk
//
// -------------------------------------------------------------------------
// DETAILS
//  Created on       : 12 Jul 20132
//  Version          : $Revision$
//  Last modified on : $Date$
//  Last modified by : $Author$
//  $HeadURL$
//
// -------------------------------------------------------------------------
// COPYRIGHT
//  Copyright (c) The University of Manchester, 2012. All rights reserved.
//  SpiNNaker Project
//  Advanced Processor Technologies Group
//  School of Computer Science
// -------------------------------------------------------------------------
// TODO
//  * everything
// -------------------------------------------------------------------------


// ----------------------------------------------------------------
// include spiNNlink global constants and parameters
//
//#`include "spiNNlink_definitions.v"
// ----------------------------------------------------------------

`timescale 1ns / 1ps
module cel
(
  input  wire rst,

  input  wire a,
  input  wire b,
  output reg  o
);
  reg r;
  reg s;

  always @(*)
    r = (!a && !b) || rst;

  always @(*)
    s = a && b;

//#  always @(posedge s or posedge r)
  always @(s or r)
    if (r)
      o <= 1'b0;
    else
      if (s)
        o <= 1'b1;
endmodule


`timescale 1ns / 1ps
module toggle
(
  input  wire rst,
  input  wire i,
  output reg  q0,
  output reg  q1
);
  always @(posedge i or posedge rst)
    if (rst)
      q0 <= 1'b0;
    else
      q0 <= ~q0;

  always @(negedge i or posedge rst)
    if (rst)
      q1 <= 1'b0;
    else
      q1 <= ~q1;
endmodule


`timescale 1ns / 1ps
module sel
(
  input  wire ri,
  output wire ai,
  output wire ro,
  input  wire ao
);
  wire i;

  cel c0
  (
    .rst (1'b0),  // no need for a reset!
    .a   (ri), 
    .b   (ao),
    .o   (i)
  );

  assign ro = ~i & ri;

  assign ai = ~(~i | ao);
endmodule


`timescale 1ns / 1ps
module seq
(
  input  wire ri,
  output wire ai,
  output wire r0,
  input  wire a0,
  output wire r1,
  input  wire a1
);
  sel s0
  (
    .ri  (ri), 
    .ai  (r1),
    .ro  (r0),
    .ao  (a0)
  );

  assign ai = a1;
endmodule


`timescale 1ns / 1ps
module platch
#(
  parameter SIZE = 1
)
(
  input  wire              rst,

  input  wire              e,
  input  wire [SIZE - 1:0] d,
  output reg  [SIZE - 1:0] q
);
//#  always @ (e or rst)
  always @ (posedge e or posedge rst)
    if (rst)
      q <= 0;
    else
      if (e)
	q <= d;
endmodule


`timescale 1ns / 1ps
module nrz_to_rtz
#(
  parameter SIZE = 1
)
(
  input  wire              rst,

  input  wire              e,
  input  wire [SIZE - 1:0] d,
  output wire [SIZE - 1:0] q
);
  reg  [SIZE - 1:0] nd;
  reg  [SIZE - 1:0] od;

  always @ (e or d or rst)
    if (rst)
      nd <= d;
    else
      if (!e)
	nd <= d;  // latch incoming data

  always @ (e or d or nd or rst)
    if (rst)
      od <= d;
    else
      if (e)
	od <= nd;  // remember incoming data

  assign q = nd ^ od;  // compute rtz data
endmodule


module decode_2of7
(
  input  wire [6:0] din,
  output reg  [3:0] dout
);
  //---------------------------------------------------------------
  // 2-of-7 decoder
  //---------------------------------------------------------------
  always @(*)
    case (din)
      7'b0010001: dout = 0;    // 0
      7'b0010010: dout = 1;    // 1
      7'b0010100: dout = 2;    // 2
      7'b0011000: dout = 3;    // 3
      7'b0100001: dout = 4;    // 4
      7'b0100010: dout = 5;    // 5
      7'b0100100: dout = 6;    // 6
      7'b0101000: dout = 7;    // 7
      7'b1000001: dout = 8;    // 8
      7'b1000010: dout = 9;    // 9
      7'b1000100: dout = 10;   // 10
      7'b1001000: dout = 11;   // 11
      7'b0000011: dout = 12;   // 12
      7'b0000110: dout = 13;   // 13
      7'b0001100: dout = 14;   // 14
      7'b0001001: dout = 15;   // 15
      7'b1100000: dout = 0;    // eop
      default:    dout = 4;    // incomplete, bad
    endcase
endmodule


module d_cd_req
(
  input  wire [6:0] din,
  input  wire       req,
  output reg        cd
);
  //---------------------------------------------------------------
  // 2of7 data completion detection with request signal
  //---------------------------------------------------------------
  always @(*)
    case ({req, din})
      8'b10010001: cd = 1'b1;   // 0
      8'b10010010: cd = 1'b1;   // 1
      8'b10010100: cd = 1'b1;   // 2
      8'b10011000: cd = 1'b1;   // 3
      8'b10100001: cd = 1'b1;   // 4
      8'b10100010: cd = 1'b1;   // 5
      8'b10100100: cd = 1'b1;   // 6
      8'b10101000: cd = 1'b1;   // 7
      8'b11000001: cd = 1'b1;   // 8
      8'b11000010: cd = 1'b1;   // 9
      8'b11000100: cd = 1'b1;   // 10
      8'b11001000: cd = 1'b1;   // 11
      8'b10000011: cd = 1'b1;   // 12
      8'b10000110: cd = 1'b1;   // 13
      8'b10001100: cd = 1'b1;   // 14
      8'b10001001: cd = 1'b1;   // 15
      default:     cd = 1'b0;   // eop, incomplete, bad
    endcase
endmodule


module de_cd_req
(
  input  wire [6:0] din,
  input  wire       req,
  output reg        cd
);
  //---------------------------------------------------------------
  // 2of7 data and end-of-packet completion detection with request signal
  //---------------------------------------------------------------
  always @(*)
    case ({req, din})
      8'b10010001: cd = 1'b1;   // 0
      8'b10010010: cd = 1'b1;   // 1
      8'b10010100: cd = 1'b1;   // 2
      8'b10011000: cd = 1'b1;   // 3
      8'b10100001: cd = 1'b1;   // 4
      8'b10100010: cd = 1'b1;   // 5
      8'b10100100: cd = 1'b1;   // 6
      8'b10101000: cd = 1'b1;   // 7
      8'b11000001: cd = 1'b1;   // 8
      8'b11000010: cd = 1'b1;   // 9
      8'b11000100: cd = 1'b1;   // 10
      8'b11001000: cd = 1'b1;   // 11
      8'b10000011: cd = 1'b1;   // 12
      8'b10000110: cd = 1'b1;   // 13
      8'b10001100: cd = 1'b1;   // 14
      8'b10001001: cd = 1'b1;   // 15
      8'b11100000: cd = 1'b1;   // eop
      default:     cd = 1'b0;   // incomplete, bad
    endcase
endmodule


module e_cd_req
(
  input  wire [6:0] din,
  input  wire       req,
  output reg        cd
);
  //---------------------------------------------------------------
  // 2of7 end-of-packet completion detection with request signal
  //---------------------------------------------------------------
  always @(*)
    case ({req, din})
      8'b11100000: cd = 1'b1;   // eop
      default:     cd = 1'b0;   // data, incomplete, bad
    endcase
endmodule


module d_cd
(
  input  wire [6:0] din,
  output reg  cd
);
  //---------------------------------------------------------------
  // 2-of-7 data completion detect
  //---------------------------------------------------------------
  always @(din)
    case (din)
      7'b0010001, // 0
      7'b0010010, // 1
      7'b0010100, // 2
      7'b0011000, // 3
      7'b0100001, // 4
      7'b0100010, // 5
      7'b0100100, // 6
      7'b0101000, // 7
      7'b1000001, // 8
      7'b1000010, // 9
      7'b1000100, // 10
      7'b1001000, // 11
      7'b0000011, // 12
      7'b0000110, // 13
      7'b0001100, // 14
      7'b0001001: // 15
//#      7'b1100000: // eop
                  cd = 1'b1;
      default   : cd = 1'b0; // incomplete or eop
    endcase
endmodule


module de_cd
(
  input  wire [6:0] din,
  output reg  cd
);
  //---------------------------------------------------------------
  // 2-of-7 data and end-of-packet completion detect
  //---------------------------------------------------------------
  always @(din)
    case (din)
      7'b0010001, // 0
      7'b0010010, // 1
      7'b0010100, // 2
      7'b0011000, // 3
      7'b0100001, // 4
      7'b0100010, // 5
      7'b0100100, // 6
      7'b0101000, // 7
      7'b1000001, // 8
      7'b1000010, // 9
      7'b1000100, // 10
      7'b1001000, // 11
      7'b0000011, // 12
      7'b0000110, // 13
      7'b0001100, // 14
      7'b0001001, // 15
      7'b1100000: // eop
                  cd = 1'b1;
      default   : cd = 1'b0; // incomplete
    endcase
endmodule


module e_cd
(
  input  wire [6:0] din,
  output reg  eop
);
  //---------------------------------------------------------------
  // 2-of-7 eop detect
  //---------------------------------------------------------------
  always @(din)
    case (din)
      7'b1100000: // EOP
                  eop = 1'b1;
      default:    eop = 1'b0; // not eop
    endcase
endmodule


module es_cd
(
  input  wire [6:0] din,
  output reg  es
);
  //---------------------------------------------------------------
  // 2-of-7 data and end-of-packet completion detect
  //---------------------------------------------------------------
  always @(din)
    case (din)
      7'b0000000, // no transitions
      7'b0000001, // 1 transition
      7'b0000010, // 1 transition
      7'b0000100, // 1 transition
      7'b0001000, // 1 transition
      7'b0010000, // 1 transition
      7'b0100000, // 1 transition
      7'b1000000, // 1 transition
      7'b0010001, // 0
      7'b0010010, // 1
      7'b0010100, // 2
      7'b0011000, // 3
      7'b0100001, // 4
      7'b0100010, // 5
      7'b0100100, // 6
      7'b0101000, // 7
      7'b1000001, // 8
      7'b1000010, // 9
      7'b1000100, // 10
      7'b1001000, // 11
      7'b0000011, // 12
      7'b0000110, // 13
      7'b0001100, // 14
      7'b0001001, // 15
      7'b1100000: // eop
                  es = 1'b0;
      default:    es = 1'b1; // error!
    endcase
endmodule


`timescale 1ns / 1ps
module Input2
(
  input  wire       rst,

  input  wire [6:0] SPLI,
  output wire       ACK,

  output wire [8:0] DATA_OUT,
  output wire       DAV,
  input  wire       DACK
);
  //---------------------------------------------------------------
  // internal signals
  //---------------------------------------------------------------
  wire	     a0;
  wire	     ia0;

  wire [6:0] i1;
  wire	     k1;

  wire	     es;
  wire	     esa;

  wire [6:0] id1;
  wire       ir1;
  wire	     ia1;
  wire	     ik1;

  wire [6:0] d1;
  wire       r1;
  wire	     a1;
  wire	     e2;
  wire	     e2a;

  wire [1:0] st;

  wire [3:0] i2;
  wire	     k2;
  wire       k2a;
  wire       k20;
  wire       k20a;
  wire       k21;
  wire       k21a;
  wire       k200;
  wire       k200a;
  wire       k201;
  wire       k201a;

  wire       eop;
  wire       d2r;

  wire [8:0] d2;
  wire       r2;
  wire	     a2;

  wire [7:0] id2;
  wire       ir2;
  wire	     ia2;


  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //--------------------------- datapath --------------------------
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //-----------------------------------------------
  // first stage -- nrz to rtz conversion
  //-----------------------------------------------
  // compute rtz data
  nrz_to_rtz #(.SIZE (7)) n2r
  (
     .rst (rst),
     .e   (ia0),
     .d   (SPLI),
     .q   (i1)
  );

  // completion detection on rtz data
  de_cd fcd1
  (
    .din (i1), 
    .cd  (k1)
  );

  // latch rtz data
  platch #(.SIZE (7)) ip1
  (
    .rst (rst),
    .e   (k1), 
    .d   (i1),
    .q   (id1)
  );

  // generate req for next stage
  cel ip1hs
  (
    .rst (rst),
    .a   (k1), 
    .b   (~ia1),
    .o   (ir1)
  );

  // input data error detection
  es_cd esd1
  (
    .din (i1), 
    .es  (es)
  );

  // complete handshake on error signal
  cel esd2
  (
    .rst (rst),
    .a   (es), 
    .b   (~esa),
    .o   (esa)
  );

  // generate internal ack signal
  assign ia0 = ir1 | esa;

  // generate nrz ack signal for SpiNnaker link
  toggle t0
  (
    .rst (rst),
    .i   (ia0),
    .q0  (a0),
    .q1  ()
  );

  // make sure to ack on reset!
  assign ACK = ~(a0 | rst);  //#
  //-----------------------------------------------

  //-----------------------------------------------
  // decoupling stage
  //-----------------------------------------------
  // check data flit complete
/*  de_cd_req ifcd1
  (
    .din (id1), 
    .req (ir1), 
    .cd  (ik1)
  );*/
  cel iifcd1
  (
    .rst (rst),
    .a   (ir1), 
    .b   (~ik1),
    .o   (ik1)
  );

  // latch rtz data
  platch #(.SIZE (7)) p1
  (
    .rst (rst),
    .e   (ik1), 
    .d   (id1),
    .q   (d1)
  );

  // generate req for next stage
  cel p1hs
  (
    .rst (rst),
    .a   (ik1), 
    .b   (~a1),
    .o   (r1)
  );

  // generate ack to previuos stage
  assign ia1 = r1;
  //-----------------------------------------------

  //-----------------------------------------------
  // second stage -- 2-of-7 decoding
  //-----------------------------------------------
  // check data flit complete
  d_cd_req cd1
  (
    .din (d1), 
    .req (r1), 
    .cd  (k2)
  );

  // check eop complete
  e_cd_req cde
  (
    .din (d1), 
    .req (r1), 
    .cd  (e2)
  );

  // generate ack signals for previous stage
  assign a1 = k2a | e2a;  //#

  //-----------------------------------------------
  // data flit
  //-----------------------------------------------
  // decode 2-of-7 signal to binary
  decode_2of7 dec
  (
    .din  (d1), 
    .dout (i2)
  );

  // de-serialise into bytes
  // de-serialiser control
  // sequence latching and updating st
  seq s2
  (
    .ri (k2),
    .ai (k2a),
    .r0 (k20),
    .a0 (k20a),
    .r1 (k21),
    .a1 (k21a)

  );

  // select latch
  assign k200 = k20 & ~st[0];
  assign k201 = k20 &  st[0];
  assign k20a = k200a | k201a;

  assign k201a = d2r;
  
  // st keeps track of which flit is received
  toggle t2
  (
//#    .rst (rst | e1),  // for safety, eop resets the state!
    .rst (rst | e2a),
//#    .rst (rst),
    .i   (k21),
    .q0  (st[0]),
    .q1  (st[1])
  );

  assign k21a = st[0] ^ st[1];  //#

  // de-serialiser datapath
  // latch first flit of decoded data
  platch #(.SIZE (4)) ip2_0
  (
    .rst (rst),
    .e   (k200), 
    .d   (i2),
    .q   (id2[3:0])
  );

  // complete handshake
  cel ip2hs_0
  (
    .rst (rst),
    .a   (k200), 
    .b   (~k200a),
    .o   (k200a)
  );

  // latch second flit of decoded data
  platch #(.SIZE (4)) ip2_1
  (
    .rst (rst),
    .e   (k201), 
    .d   (i2),
    .q   (id2[7:4])
  );

  // complete data handshake
  cel ip2hs_1
  (
    .rst (rst),
    .a   (k201), 
    .b   (~ia2),
    .o   (d2r)
  );
  //-----------------------------------------------

  //-----------------------------------------------
  // eop flit
  //-----------------------------------------------
/*//#  sel s0
  (
    .ri  (e2), 
    .ai  (e2a),
    .ro  (e20),
    .ao  (e20a)
  );
  wire	     e20;
  wire	     e20a;
  wire       e21;
  wire       e21a;
  wire [1:0] et;
   
  seq s2xxx
  (
    .ri (e2),
    .ai (e2a),
    .r0 (e20),
    .a0 (e20a),
    .r1 (e21),
    .a1 (e21a)

  );

  toggle t2xxx
  (
    .rst (rst),
    .i   (e21),
    .q0  (et[0]),
    .q1  (et[1])
  );

  assign e21a = et[0] ^ et[1];  //#
//#  assign e20 = e2;    //# testing without seq
//#  assign e2a = e20a;  //# testing without seq

//#  assign e20a = eop;
*/ //#

  // latch eop and complete handshake
  cel eophs
  (
    .rst (rst),
    .a   (e2), 
//#    .a   (e20), 
    .b   (~ia2),
    .o   (eop)
  );

  assign e2a = eop;

  assign ir2 = d2r | eop;
  //-----------------------------------------------  


  //-----------------------------------------------
  // interface to netxt stage (output!)
  //-----------------------------------------------
  // latch rtz data
  platch #(.SIZE (9)) p2
  (
    .rst (rst),
    .e   (ir2), 
    .d   ({eop, id2}),
    .q   (d2)
  );

  // complete handshake
  cel p2hs
  (
    .rst (rst),
    .a   (ir2), 
    .b   (~a2),
    .o   (r2)
  );

  assign ia2 = r2;

  assign DATA_OUT = d2;
  assign DAV      = r2;
  assign a2       = DACK;
  //-----------------------------------------------
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
endmodule
