// -------------------------------------------------------------------------
// $Id$
//  spiNNaker link driver module
//
// -------------------------------------------------------------------------
// AUTHOR
//  lap - luis.plana@manchester.ac.uk
//
// -------------------------------------------------------------------------
// DETAILS
//  Created on       : 12 Jul 20132
//  Version          : $Revision$
//  Last modified on : $Date$
//  Last modified by : $Author$
//  $HeadURL$
//
// -------------------------------------------------------------------------
// COPYRIGHT
//  Copyright (c) The University of Manchester, 2012. All rights reserved.
//  SpiNNaker Project
//  Advanced Processor Technologies Group
//  School of Computer Science
// -------------------------------------------------------------------------
// TODO
//  * everything
// -------------------------------------------------------------------------


// ----------------------------------------------------------------
// include spiNNlink global constants and parameters
//
//#`include "spiNNlink_definitions.v"
// ----------------------------------------------------------------

`timescale 1ns / 1ps
module main_tb
(
);

  //---------------------------------------------------------------
  // constants
  //---------------------------------------------------------------
  localparam RST_WIDTH = 100000;

  localparam FLT_HSDLY = 100;
  localparam SPL_HSDLY = 20;
 

  //---------------------------------------------------------------
  // internal signals
  //---------------------------------------------------------------
  reg  [8:0] DATA_IN;
  reg        DATA_IN_AV;
  wire       DATA_IN_BSY;

  wire [8:0] DATA_OUT;
  wire       DAV;
  reg        DACK;

  wire [6:0] SPLI;
  wire       ACKI; 
   

  wire [6:0] spl_data;
  reg        spl_ack; 


  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //--------------------------- functions -------------------------
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //---------------------------------------------------------------
  // NRZ 2-of-7 eop detection
  //---------------------------------------------------------------
  function [3:0] eop_nrz_2of7 ;
    input [6:0] data;
    input [6:0] old_data;

    case (data ^ old_data)
      7'b1100000:   // eop
                  eop_nrz_2of7 = 1;
      default:    eop_nrz_2of7 = 0;
    endcase
  endfunction
  //---------------------------------------------------------------

  //---------------------------------------------------------------
  // NRZ 2-of-7 decoder
  //---------------------------------------------------------------
  function [3:0] decode_nrz_2of7 ;
    input [6:0] data;
    input [6:0] old_data;

    case (data ^ old_data)
      7'b0010001: decode_nrz_2of7 = 0;    // 0
      7'b0010010: decode_nrz_2of7 = 1;    // 1
      7'b0010100: decode_nrz_2of7 = 2;    // 2
      7'b0011000: decode_nrz_2of7 = 3;    // 3
      7'b0100001: decode_nrz_2of7 = 4;    // 4
      7'b0100010: decode_nrz_2of7 = 5;    // 5
      7'b0100100: decode_nrz_2of7 = 6;    // 6
      7'b0101000: decode_nrz_2of7 = 7;    // 7
      7'b1000001: decode_nrz_2of7 = 8;    // 8
      7'b1000010: decode_nrz_2of7 = 9;    // 9
      7'b1000100: decode_nrz_2of7 = 10;   // 10
      7'b1001000: decode_nrz_2of7 = 11;   // 11
      7'b0000011: decode_nrz_2of7 = 12;   // 12
      7'b0000110: decode_nrz_2of7 = 13;   // 13
      7'b0001100: decode_nrz_2of7 = 14;   // 14
      7'b0001001: decode_nrz_2of7 = 15;   // 15
      default:    decode_nrz_2of7 = 4'hx; // eop, incomplete, oob
    endcase
  endfunction
  //---------------------------------------------------------------

  //---------------------------------------------------------------
  // NRZ 2-of-7 completion detection
  //---------------------------------------------------------------
  function [3:0] complete_nrz_2of7 ;
    input [6:0] data;
    input [6:0] old_data;

    case (data ^ old_data)
      7'b0010001,   // 0
      7'b0010010,   // 1
      7'b0010100,   // 2
      7'b0011000,   // 3
      7'b0100001,   // 4
      7'b0100010,   // 5
      7'b0100100,   // 6
      7'b0101000,   // 7
      7'b1000001,   // 8
      7'b1000010,   // 9
      7'b1000100,   // 10
      7'b1001000,   // 11
      7'b0000011,   // 12
      7'b0000110,   // 13
      7'b0001100,   // 14
      7'b0001001,   // 15
      7'b1100000:   // eop
                  complete_nrz_2of7 = 1;
      default:    complete_nrz_2of7 = 0;
    endcase
  endfunction
  //---------------------------------------------------------------

  //---------------------------------------------------------------
  // RTZ 2-of-7 decoder
  //---------------------------------------------------------------
  function [3:0] decode_2of7 ;
    input [6:0] data;

    case (data)
      7'b0010001: decode_2of7 = 0;    // 0
      7'b0010010: decode_2of7 = 1;    // 1
      7'b0010100: decode_2of7 = 2;    // 2
      7'b0011000: decode_2of7 = 3;    // 3
      7'b0100001: decode_2of7 = 4;    // 4
      7'b0100010: decode_2of7 = 5;    // 5
      7'b0100100: decode_2of7 = 6;    // 6
      7'b0101000: decode_2of7 = 7;    // 7
      7'b1000001: decode_2of7 = 8;    // 8
      7'b1000010: decode_2of7 = 9;    // 9
      7'b1000100: decode_2of7 = 10;   // 10
      7'b1001000: decode_2of7 = 11;   // 11
      7'b0000011: decode_2of7 = 12;   // 12
      7'b0000110: decode_2of7 = 13;   // 13
      7'b0001100: decode_2of7 = 14;   // 14
      7'b0001001: decode_2of7 = 15;   // 15
      default:    decode_2of7 = 4'hx; // eop, incomplete, oob
    endcase
  endfunction
  //---------------------------------------------------------------

  //---------------------------------------------------------------
  // RTZ 2-of-7 completion detection
  //---------------------------------------------------------------
  function [3:0] complete_2of7 ;
    input [6:0] data;
    input [6:0] old_data;

    case (data ^ old_data)
      7'b0010001,   // 0
      7'b0010010,   // 1
      7'b0010100,   // 2
      7'b0011000,   // 3
      7'b0100001,   // 4
      7'b0100010,   // 5
      7'b0100100,   // 6
      7'b0101000,   // 7
      7'b1000001,   // 8
      7'b1000010,   // 9
      7'b1000100,   // 10
      7'b1001000,   // 11
      7'b0000011,   // 12
      7'b0000110,   // 13
      7'b0001100,   // 14
      7'b0001001,   // 15
      7'b1100000:   // eop
                  complete_2of7 = 1;
      default:    complete_2of7 = 0;
    endcase
  endfunction
  //---------------------------------------------------------------

  //---------------------------------------------------------------
  // RTZ 2-of-7 enoder
  //---------------------------------------------------------------
  function [6:0] encode_2of7 ;
    input [4:0] data;

    casex (data)
      5'b00000 : encode_2of7 = 7'b0010001; // 0
      5'b00001 : encode_2of7 = 7'b0010010; // 1
      5'b00010 : encode_2of7 = 7'b0010100; // 2
      5'b00011 : encode_2of7 = 7'b0011000; // 3
      5'b00100 : encode_2of7 = 7'b0100001; // 4
      5'b00101 : encode_2of7 = 7'b0100010; // 5
      5'b00110 : encode_2of7 = 7'b0100100; // 6
      5'b00111 : encode_2of7 = 7'b0101000; // 7
      5'b01000 : encode_2of7 = 7'b1000001; // 8
      5'b01001 : encode_2of7 = 7'b1000010; // 9
      5'b01010 : encode_2of7 = 7'b1000100; // 10
      5'b01011 : encode_2of7 = 7'b1001000; // 11
      5'b01100 : encode_2of7 = 7'b0000011; // 12
      5'b01101 : encode_2of7 = 7'b0000110; // 13
      5'b01110 : encode_2of7 = 7'b0001100; // 14
      5'b01111 : encode_2of7 = 7'b0001001; // 15
      5'b1xxxx : encode_2of7 = 7'b1100000; // EOP
      default  : encode_2of7 = 7'bxxxxxxx;
    endcase
  endfunction
  //---------------------------------------------------------------
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++


  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //--------------------------- datapath --------------------------
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
  //--------------------------------------------------
  // device under test
  //--------------------------------------------------
  main dut
  (
    .DATA_IN           (DATA_IN),
    .DATA_IN_AV        (DATA_IN_AV),
    .DATA_IN_BSY       (DATA_IN_BSY),

    .SPIN_LINK_IN      (spl_data),
    .SPIN_LINK_IN_ACK  (spl_ack),

    .DATA_OUT          (DATA_OUT),
    .DATA_OUT_AV       (DAV),
    .DATA_O_ACK        (DACK),

    .SPIN_LINK_OUT     (SPLI),
    .SPIN_LINK_OUT_ACK (ACKI)
  );
  //--------------------------------------------------

  //--------------------------------------------------
  // drive the output flit interface
  //--------------------------------------------------
  initial
  begin
    DACK = 0;

//#    wait (rst);
//#    wait (!rst);

    forever
    begin
      wait (DAV);
      # FLT_HSDLY
        DACK = 1;

      wait (!DAV);
      # FLT_HSDLY
        DACK = 0;
    end
  end
  //--------------------------------------------------

  //--------------------------------------------------
  // drive the SpiNNaker input interface
  //--------------------------------------------------
  reg  [3:0] snd_data;
  reg        eop;
  reg  [6:0] old_spld;
  reg        old_ack;
  reg  [6:0] go;


  assign SPLI = go & (old_spld ^ encode_2of7 ({eop, snd_data}));

  initial
  begin
    go       = 7'b0000000;
    snd_data = 0;
    eop      = 0;
    old_spld = 0;

//#    wait (rst);
//#    wait (!rst);
    # (RST_WIDTH + SPL_HSDLY)
      go = 7'b1111111;

    forever
    begin
      old_ack = ACKI;
      wait (ACKI != old_ack);
      # SPL_HSDLY
        old_spld = SPLI;

      if (snd_data == 15)
      begin
        eop = 1;
        old_ack = ACKI;
        wait (ACKI != old_ack);
        # SPL_HSDLY
          old_spld = SPLI;
        eop = 0;
      end

      snd_data = snd_data + 1;
    end
  end
  //--------------------------------------------------

  //--------------------------------------------------
  // drive the input flit interface
  //--------------------------------------------------
  reg  [3:0] flt_data0;
  reg  [3:0] flt_data1;

  initial
  begin
    flt_data0 = 4'd0;
    flt_data1 = 4'd1;

    DATA_IN = 9'b111111111; // reset the cpld!
    DATA_IN_AV = 1'b0;

//#    wait (rst);
//#    wait (!rst);
    # RST_WIDTH
      DATA_IN = 9'b000000000;

    forever
    begin
      # FLT_HSDLY 
        wait (~DATA_IN_BSY);
      if (flt_data0 == 14)
      begin
        DATA_IN = 9'b100000000; // send eop
        #5
          DATA_IN_AV = ~DATA_IN_AV;
        flt_data0 = 4'd0;
        flt_data1 = 4'd1;
      end
      else
      begin
        DATA_IN = {1'b0, flt_data1, flt_data0};  // send next flit
        #5
          DATA_IN_AV = ~DATA_IN_AV;
        flt_data0 = flt_data0 + 2;
        flt_data1 = flt_data1 + 2;
      end
    end
  end
  //--------------------------------------------------

  //--------------------------------------------------
  // drive the output SpiNNaker interface
  //--------------------------------------------------
  reg  [6:0] old_data;
  wire [3:0] rec_data;
  wire       rec_eop;

  assign rec_data = decode_nrz_2of7 (spl_data, old_data);
  assign rec_eop  = eop_nrz_2of7 (spl_data, old_data);

  initial
  begin
    spl_ack = 1'b0;
    old_data = 7'd0;

//#    wait (rst);
//#    wait (!rst);

    # (RST_WIDTH + SPL_HSDLY)
//#      spl_ack = 1'b1;
    old_data = spl_data;

    forever
    begin
      wait (complete_nrz_2of7 (spl_data, old_data));
      # SPL_HSDLY
        spl_ack = ~spl_ack;
      old_data = spl_data;
    end
  end
  //--------------------------------------------------
  //+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
endmodule
